@@ -234,9 +234,14 @@ static const int64_t AV_CODEC_ID_VIDEO_TAB[][2] = {
   {(1719680309), AV_CODEC_ID_MWSC},
   {(1939479851), AV_CODEC_ID_WCMV},
   {(1828387903), AV_CODEC_ID_RASC},
+  {(1608980913), AV_CODEC_ID_HYMT},
+  {(1453025829), AV_CODEC_ID_ARBC},
+  {(6510927), AV_CODEC_ID_AGM},
+  {(1697096237), AV_CODEC_ID_LSCR},
+  {(8603445), AV_CODEC_ID_VP4},
 };
 
-#define AV_CODEC_ID_VIDEO_TAB_LEN 233
+#define AV_CODEC_ID_VIDEO_TAB_LEN 238
 enum AVCodecID VideoCodecID_val(value v){
 int i;
 for(i=0;i<AV_CODEC_ID_VIDEO_TAB_LEN;i++){
@@ -329,6 +334,7 @@ static const int64_t AV_CODEC_ID_AUDIO_TAB[][2] = {
   {(-1192040419), AV_CODEC_ID_ADPCM_AICA},
   {(-1387207025), AV_CODEC_ID_ADPCM_IMA_DAT4},
   {(-924797655), AV_CODEC_ID_ADPCM_MTAF},
+  {(1593228927), AV_CODEC_ID_ADPCM_AGM},
   {(386199643), AV_CODEC_ID_AMR_NB},
   {(386203657), AV_CODEC_ID_AMR_WB},
   {(-1004210045), AV_CODEC_ID_RA_144},
@@ -428,9 +434,10 @@ static const int64_t AV_CODEC_ID_AUDIO_TAB[][2] = {
   {(-503198383), AV_CODEC_ID_APTX_HD},
   {(8298921), AV_CODEC_ID_SBC},
   {(648283377), AV_CODEC_ID_ATRAC9},
+  {(1606793715), AV_CODEC_ID_HCOM},
 };
 
-#define AV_CODEC_ID_AUDIO_TAB_LEN 176
+#define AV_CODEC_ID_AUDIO_TAB_LEN 178
 enum AVCodecID AudioCodecID_val(value v){
 int i;
 for(i=0;i<AV_CODEC_ID_AUDIO_TAB_LEN;i++){
@@ -471,9 +478,10 @@ static const int64_t AV_CODEC_ID_SUBTITLE_TAB[][2] = {
   {(6516291), AV_CODEC_ID_ASS},
   {(-1185551519), AV_CODEC_ID_HDMV_TEXT_SUBTITLE},
   {(1874633215), AV_CODEC_ID_TTML},
+  {(691667747), AV_CODEC_ID_ARIB_CAPTION},
 };
 
-#define AV_CODEC_ID_SUBTITLE_TAB_LEN 25
+#define AV_CODEC_ID_SUBTITLE_TAB_LEN 26
 enum AVCodecID SubtitleCodecID_val(value v){
 int i;
 for(i=0;i<AV_CODEC_ID_SUBTITLE_TAB_LEN;i++){
